import processing.serial.*; //Serial-Library with which i can communicate with Arduino if necessary
import processing.video.Capture; //Camera Library so that I can use a camera
import processing.video.*;  //Video-Library with which I can work with videos

Serial myPort;  //Serial-Object
String val; //String to save the Arduino-Outputs

Capture video; //Capture-Object 

int numPixels; //Amount of pixels per frame
long[] timestamp; //

int THRESHHOLD = 120; //The lower the brightness threshhold, the more white pixels there are. This means better contrast, but you also need a darker object moving throught the frame.
int counter = 0; //Counter for the amount of pictures taken
float SCALELONG = 0.94; //The distance between the lasers in meters , easily established real length
int clickCounter = 0; //Counts the clicks, necessary for establishing scale


boolean go = false;  //Code-Control boolean with which I control taking the pictures 

float finSpeed, scale, scaleUnfin, camSpeed; //Assortment of global float-variables
int start, end; //The numerical values with which I set the scale, are initialized in MousePressed

color BLACK = color(0); //Color-Variables to increase contrast of the pictures, more through explanation in normaliser() function
color WHITE = color(255);
PImage[] pics = new PImage[200]; //Picture-Array where the pictures for the velocity measurement are saved

/*In setup I initialize the Serial and Capture Objects so I can communicate with the camera and Arduino. 
 I also set the framerate to 30 and initalize the array for the times I take the different pictures. */
void setup()
{
  size(640, 480);
  String[] cameras = Capture.list();
  String portName = Serial.list()[1]; 
  myPort = new Serial(this, portName, 9600);
  video = new Capture(this, cameras[2]); //Camera-Input 
  video.start();   //Starts video
  frameRate(30);  //Sets framerate to 30, the standard
  timestamp = new long[25]; //Array for the time, up to 25 pictures, can be more
  numPixels = video.width * video.height; //Establishes amount of pixels
}

/*THe draw function is repeated again and again by Processing. Its also where input is taken from the Arduino, 
 pictures are taken, the scale is established, and the speed measured by the Arduino and camera are outputted. 
 The pictures taken are saved in the Sketch Folder of this program. */
void draw() {

  if ( myPort.available() > 0) {  
    val = myPort.readStringUntil('\n');
    //println(val); 
  }
  if ( video.available() == true) {
    video.read();
    image(video, 0, 0);
    if (clickCounter == 2) { //Sets the scale for the measurement 
      scaleUnfin = abs(end - start)/10; 
      scale = SCALELONG/scaleUnfin; 
      println("Der Masstab ist " + scale); 
      clickCounter++;
    }
    if ( (frameCount % 3 == 0 && go == true) && counter<20) { //Takes twenty pictures, one every 5 frames, or at whatever rate one wants (at least one every 5 frames though)
      saveFrame("BlitzBild-######.jpg"); //Saves the picture for later use
      pics[counter] = video.get(); //Saves the picture into an array for later speed-measurement 
      //println("Frame saved!"); Diagnosticcode
      timestamp[counter] = millis();  //Saves the time at which the picture was taken
      counter++; //Increases the counter by 1
    }

    if (counter >= 20 && go == true) { //After 20 pictures are taken the speed is calculated using them and SpeedAv and the included functions. 
      camSpeed = SpeedAv(); 
      println("Die von der Kamera gemessenen Geschwindigkeit ist " + abs(camSpeed) + "m/s bzw. " + abs(camSpeed)*3.6 + "km/h. "); //outputs the speed measured by the camera 
      println("Die von der Lichtschranke gemessenen Geschwindigkeit ist " + val + "m/s das bedeutet, das der Messfehler " + abs(float(val) - camSpeed) + " beträgt." ); 
      if(float(val) > 3){
        println("Das war zu schnell!"); 
      }
      /*Prints the speed measured by the Arduino and compares to camera-speed*/
      go = false;
    }
  }
  if (keyPressed == true) { //Starts the program when a key is pressed by 'flipping switch' on pictures being taken 
    go = true;
  }
}



/*In edgefinder() the edge of a dark object on a picture is found by comparing pixels to their neighbors
 and saving the positions of the pixels where that value is over a certain threshhold. Also treats the pixel array like a grid wih x and y values for this purpose. */
float edgefinder(PImage im1) { 
  normalizer(im1);  //Increases the light-dark contrast to make the edge clearer 
  IntList edge = new IntList(); //Saves the x-values of the edge in a list 
  for (int y = 0; y < im1.height; y++) { //Goes through all y-values 
    for (int x = 1; x < im1.width; x++ ) { //Goes through all x-Values 
      //The location of the current pixel, ie its position in the array 
      int loc = x + y*im1.width;
      int pix = im1.pixels[loc];
      //The location of the previous pixel, ie its position in the array  
      int leftLoc = (x-1) + y*im1.width;
      int leftPix = im1.pixels[leftLoc];
      //Calculates the brightness-difference between the two pixels 
      float diff = abs(brightness(pix) - brightness(leftPix));
      //Checks if the brightnessdifference is above the brightness value, which it should always be at an edge thanks to normalizer 
      if (diff > THRESHHOLD) { 
        edge.append(x); //If the difference big enough then the x-Value of that pixel is added to the list.
      }
    }
  }
  float edgeAv1 = 0; 
  int j = 0; 
  while ( j < ((edge.size())-1)) { //Iterates through the list of all the x-values for edge pixels and makes the average x-value
    edgeAv1 += edge.get(j);
    j++;
  }
  //println(j); Diagnostic code
  if ( j!= 0) { //Checks for j=0, because there we would get a NaN error; 
    edgeAv1 = edgeAv1/j; 
    if (edgeAv1 > 0 && edgeAv1 != Float.NaN) {
      println("The first edge is at " + edgeAv1); 
      return edgeAv1 ;
    } else { //Everything after this is mainly diagnostic code and throws exception 
      //println("Edge not found uh oh!!");
    }
  }

  // println("wonky value!"); 
  return 99999;//Error value that cant be acheived normally because the width is lower than this number
}

/*Calculates the edge in a second pictrue and subtracts the x-Values of the two edges to get the distance covered by the object in between those pictures.
USes same algorithm as edgefinder */
float dist(PImage im1, PImage im2) { 
  float edgeAv1 = edgefinder(im1); 
  normalizer(im2); 
  float edgeAv2 = 0;  

  IntList edge2 = new IntList();
  for (int y = 0; y < im2.height; y++) {
    for (int x = 1; x < im2.width; x++ ) {

      int loc = x + y*im2.width;
      color pix = im2.pixels[loc];


      int leftLoc = (x-1) + y*im2.width;
      color leftPix = im2.pixels[leftLoc];


      float diff = abs(brightness(pix) - brightness(leftPix));
      if (diff > THRESHHOLD) {
        edge2.append(x);
      }
    }
  }

  int j = 0; 
  while ( j < ((edge2.size())-1)) {
    edgeAv2 += edge2.get(j);
    j++;
  }
  //println(j); Good diagnostic tool
  if (j != 0 && edgeAv1 != Float.NaN) { //Checks, that no nonsense values are produced do to faulty edge-capture
    edgeAv2 = edgeAv2/j; 
    println("The 2nd edge is at " + edgeAv2); 
    float dist = ((float)(edgeAv2 - edgeAv1));
    println(dist);
    if( dist < 0){ return 0;}
    else{
    return dist;
    }
  } else {
    println("edge not found!! ");
    println("wonky value!"); 
    return 99999.0; //Mistake value, that cannot be normally reached by the program
  }
}


float Speed(PImage im1, PImage im2, int i) { //Calculates the velocity by: v=s/t
  float distance = dist(im1, im2); 
  if (distance == 99999) {
    return 99999;
  } else {
    float distReal = scale * distance; 
    float time = (timestamp[i+1] - timestamp[i]); //Timespan between capture of pictures
    //println(time); 
    float speed = (distReal/time)*1000; //Um von ms zu s umzuwechseln
    //println("The speed is " + speed ); 
    return speed;
  }
}
//calculates the average velocity between all pictures using Speed()
float SpeedAv() {
  float speedAv = 0; 
  int i = 0; 
  int ii = 0; 
  int numPic = 15; //Amount of pictures over which speed should be calculated, needs to be smaller than the amount of pictures taken  
  while(i < numPic) { 
    if (Speed(pics[i], pics[i+1], i) != 99999) { //Checks if the error value of the speed function (ie where no edge was found in the picture) is returned, if so that value is skipped
      speedAv += Speed(pics[i], pics[i+1], i); //Adds all the "real" speed values together by using Speed()
      ii++; 
    }
    i++; 
  }
  finSpeed = speedAv/ii; //Divide by the number of pictures to get the average
  return finSpeed; //returns the final calculated speed
}

//Increases light-dark contrast, makes finding the edge easier as it reduces gradient
void normalizer( PImage image) { 
  int i = 0;  
  float pixelBrightness; 
  for ( i = 0; i < numPixels; i++) { //Goes through every pixel of the picture 
    pixelBrightness = brightness(image.pixels[i]); //Checks brightness of the pixel 
    if (pixelBrightness > THRESHHOLD) { 
      image.pixels[i] = WHITE; //If its brighter than the threshold it turns white
    } else { 
      image.pixels[i] = BLACK; //if the pixel is darker than the threshhold it is set to black
    }
  }
}

void mousePressed() { //Establishes scale by clicking on the ends of an object whose length we know ie in this case our laser-timer contraption
  if (clickCounter == 0) {
    start = mouseX; //X-value of the first click 
    clickCounter++;
  } else if (clickCounter == 1) {
    end = mouseX;  //X-Value of the second click 
    clickCounter++;
  }
}
